import datetime

import pytest
from django.core.exceptions import ValidationError
from django.db import IntegrityError
from obapi import utils
from obapi.models import Author, ContentItem, Idea, Tag, Topic
from obapi.models.classifiers import CLASSIFIER_SLUG_MAX_LENGTH, IdeaAlias, TopicAlias


@pytest.mark.django_db
class TestCreateAliasedModel:
    def test_create_valid_topic(self):
        # Act
        Topic.objects.create(name="Law&Øther", description="Law and other topics.")

        # Assert
        obj = Topic.objects.get(name="Law&Øther")
        assert obj.name == "Law&Øther"
        assert obj.slug == "law-other"
        assert obj.description == "Law and other topics."

        obj_aliases = obj.aliases.all()
        assert obj_aliases.count() == 1

        alias = obj_aliases.get()
        assert alias.text == "law-other"
        assert alias.protected

    def test_create_valid_topic_with_aliases(self):
        # Act
        topic = Topic.objects.create(name="topic-two")
        topic.aliases.create(text="Second Topic")

        # Assert
        obj = Topic.objects.get(name="topic-two")
        assert obj.name == "topic-two"
        assert obj.slug == obj.name
        assert obj.description == ""

        obj_aliases = obj.aliases.all()
        assert obj_aliases.count() == 2

        expected_aliases = {"topic-two", "second-topic"}
        actual_aliases = set(obj_aliases.values_list("text", flat=True))
        assert actual_aliases == expected_aliases

        name_alias = obj_aliases.get(text="topic-two")
        assert name_alias.protected

    @pytest.mark.parametrize(
        "new_name,new_slug",
        [
            ("Law&Other", "law-other"),
            ("law other", "law-other"),
            ("law-other", "law-other"),
            ("LawOther", "lawother"),
        ],
    )
    def test_change_topic_name(self, new_name, new_slug):
        # Act
        example = Topic.objects.create(
            name="Law&Øther", description="Law and other topics."
        )
        example.name = new_name
        example.save()

        # Assert
        assert example.name == new_name
        assert example.slug == new_slug

    def test_can_create_another_object_type_with_same_name(self):
        # Act
        topic = Topic.objects.create(name="Duplicated", description="A duplicate.")
        idea = Idea.objects.create(name="Duplicated", description="A duplicate.")

        # Assert
        assert topic.name == idea.name
        assert topic.slug == idea.slug
        topic_alias_set = set(topic.aliases.values_list("text", flat=True))
        idea_alias_set = set(idea.aliases.values_list("text", flat=True))
        assert topic_alias_set == idea_alias_set

    def test_cannot_duplicate_aliases(self):
        # Arrange
        topic = Topic.objects.create(name="Law&Other", description="About law.")
        topic.aliases.create(text="law-etc")

        # Act & assert
        # Same name and description
        with pytest.raises(IntegrityError):
            Topic.objects.create(name="Law&Other", description="About law.")

        # Different name, same slug
        with pytest.raises(IntegrityError):
            Topic.objects.create(name="Law Other")

        # Name = existing alias
        with pytest.raises(IntegrityError):
            Topic.objects.create(name="law-etc")

        # Slug = existing alias
        with pytest.raises(IntegrityError):
            Topic.objects.create(name="Law, Etc.")


@pytest.mark.django_db
class TestCreateWithAliases:
    @pytest.mark.parametrize(
        "topic_name,aliases,alias_count",
        [
            ("Law Other", [], 1),
            ("Law Other", ["law", "norms"], 3),
            ("Law Other", ["law-other"], 1),
        ],
    )
    def test_create_valid_topics(self, topic_name, aliases, alias_count):
        # Act
        t1 = Topic.objects.create_with_aliases(name=topic_name, aliases=aliases)

        # Assert
        assert t1.name == topic_name
        assert t1.slug == utils.to_slug(
            topic_name, max_length=CLASSIFIER_SLUG_MAX_LENGTH
        )
        assert t1.aliases.count() == alias_count

    @pytest.mark.parametrize(
        "topic_name,aliases", [("Law", ["legal"]), ("Legal", ["lawyer", "norms"])]
    )
    def test_fails_with_invalid_names(self, topic_name, aliases):
        # Arrange
        Topic.objects.create_with_aliases(name="Law", aliases=["law-etc", "norms"])

        # Act & Assert
        with pytest.raises(IntegrityError):
            Topic.objects.create_with_aliases(name=topic_name, aliases=aliases)


@pytest.mark.django_db
class TestValidateUniqueAlias:
    def test_succeeds_for_unique_aliases(self):
        # Arrange
        law = Topic.objects.create(name="Law", description="About law.")
        law.aliases.create(text="legal")
        norms = Topic.objects.create(name="Norms", description="About norms.")
        social = TopicAlias(text="social-norms", owner=norms)

        # Act & assert
        try:
            social.full_clean()  # not yet saved
        except ValidationError:
            pytest.fail("ValidationError raised with unique aliases.")
        social.save()
        try:
            social.full_clean()
        except ValidationError:
            pytest.fail("ValidationError raised with unique aliases.")

    def test_succeeds_when_another_object_type_has_same_alias(self):
        # Arrange
        law_topic = Topic.objects.create(name="Law", description="About law.")
        law_topic.aliases.create(text="rules")
        norms_idea = Idea.objects.create(name="Norms", description="About norms.")
        rules = IdeaAlias(text="rules", owner=norms_idea)

        # Act & assert
        try:
            rules.full_clean()  # not yet saved
        except ValidationError:
            pytest.fail("ValidationError raised with unique aliases.")
        rules.save()
        try:
            rules.full_clean()
        except ValidationError:
            pytest.fail("ValidationError raised with unique aliases.")

    @pytest.mark.parametrize("alias_text", ["Law", "law", "Rules", "rulEs"])
    def test_fails_with_duplicate_aliases(self, alias_text):
        # Arrange
        law = Topic.objects.create(name="Law", description="About law.")
        law.aliases.create(text="Rules")
        norms = Topic.objects.create(name="Norms", description="About norms.")

        # Act & Assert
        norms_alias = TopicAlias(text=alias_text, owner=norms)
        with pytest.raises(ValidationError):
            norms_alias.full_clean()


@pytest.mark.django_db
class TestMergeObjects:
    def test_preserves_related_content_for_topics(self):
        # Arrange - create topics and related content
        law = Topic.objects.create_with_aliases(name="Law", aliases=["legal", "laws"])
        norms = Topic.objects.create_with_aliases(name="Norms", aliases=["norm"])

        now = datetime.datetime.now()
        video = ContentItem.objects.create(title="Video Item", publish_date=now)
        audio = ContentItem.objects.create(title="Audio Item", publish_date=now)
        text = ContentItem.objects.create(title="Text Item", publish_date=now)

        law.content.add(video, audio)
        norms.content.add(text)

        # Act
        merged_object = Topic.objects.all().merge_objects()

        # Assert
        assert not Topic.objects.filter(pk__in=[law.pk, norms.pk])

        expected_content = {video, audio, text}
        assert set(merged_object.content.all()) == expected_content

        expected_names = ("Law", "Norms")
        assert merged_object.name in expected_names

        expected_aliases = {"law", "legal", "laws", "norms", "norm"}
        actual_aliases = set(merged_object.aliases.values_list("text", flat=True))
        assert actual_aliases == expected_aliases

    def test_can_merge_objects_without_descriptions(self):
        # Arrange
        law = Tag.objects.create_with_aliases(name="Law", aliases=["legal", "laws"])
        norms = Tag.objects.create_with_aliases(name="Norms", aliases=["norm"])

        # Act
        merged_object = Tag.objects.all().merge_objects()

        # Assert
        assert not Tag.objects.filter(pk__in=[law.pk, norms.pk])

        expected_names = ("Law", "Norms")
        assert merged_object.name in expected_names

        expected_aliases = {"law", "legal", "laws", "norms", "norm"}
        actual_aliases = set(merged_object.aliases.values_list("text", flat=True))
        assert actual_aliases == expected_aliases

    def test_preserves_related_content_for_authors(self):
        # Arrange - create topics and related content
        jane = Author.objects.create_with_aliases(name="Jane", aliases=["janet", "jan"])
        norms = Author.objects.create_with_aliases(name="Oscar", aliases=["osc"])

        now = datetime.datetime.now()
        video = ContentItem.objects.create(title="Video Item", publish_date=now)
        audio = ContentItem.objects.create(title="Audio Item", publish_date=now)
        text = ContentItem.objects.create(title="Text Item", publish_date=now)

        jane.content.add(video, audio)
        norms.content.add(text)

        # Act
        merged_object = Author.objects.all().merge_objects()

        # Assert
        assert not Author.objects.filter(pk__in=[jane.pk, norms.pk])

        expected_content = {video, audio, text}
        assert set(merged_object.content.all()) == expected_content

        expected_names = ("Jane", "Oscar")
        assert merged_object.name in expected_names

        expected_aliases = {"jane", "janet", "jan", "oscar", "osc"}
        actual_aliases = set(merged_object.aliases.values_list("text", flat=True))
        assert actual_aliases == expected_aliases


@pytest.mark.django_db
class TestConvertObject:
    def test_preserves_related_content(self):
        # Arrange - create topic and related content
        law_topic = Topic.objects.create_with_aliases(
            name="Law", description="A short description", aliases=["legal", "laws"]
        )

        now = datetime.datetime.now()
        video = ContentItem.objects.create(title="Video Item", publish_date=now)
        audio = ContentItem.objects.create(title="Audio Item", publish_date=now)
        text = ContentItem.objects.create(title="Text Item", publish_date=now)

        law_topic.content.add(video, audio, text)

        # Act
        converted_object = law_topic.convert_object(Idea)

        # Assert
        with pytest.raises(Topic.DoesNotExist):
            law_topic.refresh_from_db()

        expected_content = {video, audio, text}
        assert set(converted_object.content.all()) == expected_content

        assert converted_object.name == "Law"
        assert converted_object.description == "A short description"

        expected_aliases = {"law", "legal", "laws"}
        actual_aliases = set(converted_object.aliases.values_list("text", flat=True))
        assert actual_aliases == expected_aliases

    def test_can_convert_topic_to_author(self):
        # Arrange
        example_topic = Topic.objects.create_with_aliases(
            name="Example", description="A short description", aliases=["test"]
        )

        # Act
        converted_object = example_topic.convert_object(Author)

        # Assert
        with pytest.raises(Topic.DoesNotExist):
            example_topic.refresh_from_db()

        assert converted_object.name == "Example"
        assert converted_object.description is None

        expected_aliases = {"example", "test"}
        actual_aliases = set(converted_object.aliases.values_list("text", flat=True))
        assert actual_aliases == expected_aliases

    @pytest.mark.parametrize(
        "topic_name,aliases",
        [("Law", []), ("law", []), ("Norms", ["law"]), ("Norms", ["legal"])],
    )
    def test_fails_when_alias_already_exists(self, topic_name, aliases):
        # Arrange
        Idea.objects.create_with_aliases(name="Law", aliases=["legal", "laws"])
        law_topic = Topic.objects.create_with_aliases(name=topic_name, aliases=aliases)

        # Act & Assert
        with pytest.raises(IntegrityError):
            law_topic.convert_object(Idea)

        try:
            Topic.objects.get(pk=law_topic.pk)
        except Topic.DoesNotExist:
            pytest.fail("Topic was deleted during failed conversion.")

    def test_can_convert_object_without_description(self):
        law_tag = Tag.objects.create_with_aliases(name="Law", aliases=["legal", "laws"])

        converted_object = law_tag.convert_object(Idea)

        # Assert
        with pytest.raises(Tag.DoesNotExist):
            law_tag.refresh_from_db()

        assert converted_object.name == "Law"
        assert converted_object.description == ""

        expected_aliases = {"law", "legal", "laws"}
        actual_aliases = set(converted_object.aliases.values_list("text", flat=True))
        assert actual_aliases == expected_aliases



from django.db import models, transaction
from django.db.models import F, Q
from django.urls import reverse
from model_utils.managers import InheritanceQuerySet
from obapi import utils
from obapi.assemble import (
    assemble_ob_content_items,
    assemble_ob_edit_dates,
    assemble_spotify_content_items,
    assemble_youtube_content_items,
)
from obapi.converters import (
    OBPostLongURLConverter,
    OBPostShortURLConverter,
    SpotifyEpisodeURLConverter,
    YoutubeVideoURLConverter,
)
from obapi.models import (
    CLASSIFIER_SLUG_MAX_LENGTH,
    Author,
    ExternalLink,
    Idea,
    Tag,
    Topic,
)


class ContentItemQuerySet(InheritanceQuerySet):
    """Custom QuerySet which can "assemble" and save objects."""

    # Converter from URLs to item IDs
    url_converters = ()

    def save_item(self, item=None, **kwargs):
        """Create a new item or update an existing item."""
        adding = item is None
        author_names = kwargs.pop("author_names", None)
        classifier_names = kwargs.pop("classifier_names", None)
        link_urls = kwargs.pop("link_urls", None)
        with transaction.atomic():
            # Update or create object
            if adding:
                item = self.create(**kwargs)
            else:
                for attr, value in kwargs.items():
                    setattr(item, attr, value)
                item.save()
            # Set ManyToMany related objects: authors, classifiers, links
            authors = self.get_or_create_authors_by_names(author_names)
            ideas, topics, tags = self.get_or_create_classifiers_by_names(
                classifier_names
            )
            external_links = self.get_or_create_external_links_by_urls(link_urls)
            relationship_attributes = {
                "authors": authors,
                "ideas": ideas,
                "topics": topics,
                "tags": tags,
                "external_links": external_links,
            }
            for attr, value in relationship_attributes.items():
                if value is not None:
                    # Clear old values, set new values
                    getattr(item, attr).set(value)
            # Only internalize links if some are added
            item.internalize_links(clear=external_links is not None)
        return item

    def get_or_create_authors_by_names(self, author_names):
        """Get or create a list of authors from some names."""
        if author_names is None:
            return None
        with transaction.atomic():
            # Match by alias
            authors = [
                Author.objects.get_or_create(
                    alias__text=utils.to_slug(
                        author_name, max_length=CLASSIFIER_SLUG_MAX_LENGTH
                    ),
                    defaults={"name": author_name},
                )
                for author_name in author_names
            ]
            return [author[0] for author in authors]

    def get_or_create_classifiers_by_names(self, classifier_names):
        """Get or create a list of ideas, topics and tags from some names."""
        if classifier_names is None:
            return None, None, None
        ideas = []
        topics = []
        tags = []
        with transaction.atomic():
            for classifier_name in classifier_names:
                # Match by alias
                query = Q(
                    alias__text=utils.to_slug(
                        classifier_name, max_length=CLASSIFIER_SLUG_MAX_LENGTH
                    )
                )
                try:
                    ideas.append(Idea.objects.get(query))
                    continue
                except Idea.DoesNotExist:
                    pass
                # If that fails, try to get topic with alias = classifier_name
                try:
                    topics.append(Topic.objects.get(query))
                    continue
                except Topic.DoesNotExist:
                    pass
                # If that fails, try to get a tag with alias = classifier_name
                # If no such tag exists, create a new one
                tags.append(
                    Tag.objects.filter(query).get_or_create(
                        defaults={"name": classifier_name},
                    )[0]
                )
        return ideas, topics, tags

    def get_or_create_external_links_by_urls(self, link_urls):
        if link_urls is None:
            return None
        with transaction.atomic():
            return [
                ExternalLink.objects.get_or_create(url=link_url)[0]
                for link_url in link_urls
            ]

    def assemble_items(self, item_ids=None):
        """Assemble items in QuerySet or from their item IDs."""
        if item_ids is None:
            # Construct item ids from QuerySet
            item_ids = [item.item_id for item in self]

        # Return if no IDs
        if item_ids == []:
            return []

        assembled_items = type(self).assemble_by_ids(item_ids)
        return assembled_items

    def create_item(self, item_id):
        """Create a single content item."""
        return self.create_items([item_id])[0]

    def create_items(self, item_ids):
        """Create items from their item IDs.

        Returns
        -------
        List[ContentItem | None]
        """
        # Assemble data
        assembled_items = self.assemble_items(item_ids)

        # Save items
        created_items = []
        with transaction.atomic():
            for item_data in assembled_items:
                if item_data is not None:
                    created_items.append(self.save_item(item=None, **item_data))
                else:
                    created_items.append(None)
        return created_items

    def update_items(self, exclude=None):
        """Update items in QuerySet, excluding certain fields.

        Returns
        -------
        List[Tuple[ContentItem, bool]]
            A list of tuples of the form (`item`, `updated`).
        """
        # Assemble data
        assembled_items = self.assemble_items()

        # Don't update excluded attributes
        if exclude is None:
            exclude = []
        for new_item in assembled_items:
            for attr in exclude:
                new_item.pop(attr, None)

        # Save items
        updated_items = []
        with transaction.atomic():
            for item, item_data in zip(self, assembled_items):
                if item_data is not None:
                    updated_item = self.save_item(item=item, **item_data)
                    updated_items.append((updated_item, True))
                else:
                    # Item assemble failed - return original item
                    updated_items.append((item, False))
        return updated_items

    def find_by_url(self, url):
        """Find a ContentItem by its URL.

        Raises
        ------
        ValueError
            If the URL does not match any recognised pattern.
        ContentItem.DoesNotExist
            If the URL matches a recognised pattern but no ContentItem is found.
        """
        # Try to use get_by_url for current class (if implemented)
        try:
            return self.get_by_url(url)
        except ValueError:
            # Don't catch Model.DoesNotExist
            pass

        # Otherwise try find_by_url methods of subclasses
        for model in self.model.__subclasses__():
            try:
                return model.objects.find_by_url(url)
            except ValueError:
                pass
            # Don't catch model.DoesNotExist
        raise ValueError  # URL did not match any ContentItem URL format

    def get_by_url(self, url):
        """Get a ContentItem by its URL.

        Raises
        ------
        ValueError
            If the URL does not match any content type.
        ContentItem.DoesNotExist
            If the URL matches a content type but no content item was found.
        """

        for converter, field in self.url_converters:
            try:
                item_id = converter.to_id(url)
            except ValueError:
                pass
            else:
                return self.get(**{field: item_id})
        raise ValueError(f"URL did not match any known format for type {type(self)}")

    def recent(self):
        """Return all content items sorted by publish date."""
        return self.order_by("-publish_date")


class ContentItem(models.Model):
    objects = ContentItemQuerySet.as_manager()

    create_timestamp = models.DateTimeField("creation date", auto_now_add=True)
    update_timestamp = models.DateTimeField("update date", auto_now=True)
    download_timestamp = models.DateTimeField(
        "download date", editable=False, null=True
    )

    title = models.CharField(max_length=100, help_text="Title of content.")
    description_html = models.TextField(
        max_length=5000, blank=True, help_text="HTML description of content."
    )
    publish_date = models.DateTimeField(help_text="Date of publication.")
    edit_date = models.DateTimeField(
        blank=True, null=True, help_text="Date of last edit."
    )
    authors = models.ManyToManyField(
        Author, related_name="content", help_text="Authors."
    )
    ideas = models.ManyToManyField(
        Idea,
        related_name="content",
        related_query_name="content",
        blank=True,
        help_text="Ideas discussed or referenced by the content.",
    )
    topics = models.ManyToManyField(
        Topic,
        related_name="content",
        related_query_name="content",
        blank=True,
        help_text="Topics discussed or references by the content.",
    )
    tags = models.ManyToManyField(
        Tag,
        related_name="content",
        related_query_name="content",
        blank=True,
        help_text="Miscellaneous tags.",
    )
    external_links = models.ManyToManyField(
        ExternalLink,
        related_name="content",
        related_query_name="content",
        blank=True,
        help_text="Links to external webpages.",
    )
    internal_links = models.ManyToManyField(
        "self",
        related_name="internal_pingbacks",
        related_query_name="internal_pingbacks",
        blank=True,
        help_text="Links to other content items.",
        symmetrical=False,
    )

    def __str__(self):
        return self.title

    @property
    def content_url(self):
        # Return URL of subclass
        return ContentItem.objects.get_subclass(pk=self.pk).content_url

    def get_absolute_url(self):
        return ContentItem.objects.get_subclass(pk=self.pk).get_absolute_url()

    def internalize_links(self, clear=False):
        """Try to make external links internal.

        Try to make external links internal by looking for the corresponding
        ContentItems.

        Parameters
        ----------
        clear : bool
            Whether to clear existing internal links.
        """
        if clear:
            self.internal_links.clear()
        for pk, url in self.external_links.values_list("pk", "url"):
            try:
                match = ContentItem.objects.find_by_url(url)
            except (ValueError, ContentItem.DoesNotExist):
                continue
            else:
                with transaction.atomic():
                    self.external_links.remove(pk)
                    self.internal_links.add(match)


class VideoContentItem(ContentItem):
    duration = models.DurationField(
        help_text="Duration of video.", blank=True, null=True
    )
    view_count = models.PositiveIntegerField(
        help_text="Number of video views.", blank=True, null=True
    )

    class Meta:
        verbose_name = "video"


class YoutubeContentItemQuerySet(ContentItemQuerySet):
    assemble_by_ids = assemble_youtube_content_items
    url_converters = ((YoutubeVideoURLConverter(), "item_id"),)


class YoutubeContentItem(VideoContentItem):
    objects = YoutubeContentItemQuerySet.as_manager()
    item_id = models.CharField(
        "video ID", max_length=30, unique=True, help_text="Video ID."
    )
    yt_channel_id = models.CharField(
        "channel ID", max_length=30, help_text="Channel ID."
    )
    yt_channel_title = models.CharField(
        "channel title", max_length=30, help_text="Channel title."
    )
    yt_likes = models.PositiveIntegerField(
        "likes", blank=True, null=True, help_text="Number of likes."
    )
    yt_description = models.TextField(
        "description", max_length=5000, blank=True, help_text="Video description."
    )

    @property
    def content_url(self):
        return YoutubeVideoURLConverter().to_url(self.item_id)

    def get_absolute_url(self):
        return reverse("youtubecontentitem_detail", kwargs={"item_id": self.item_id})

    @property
    def site_url(self):
        return "https://www.youtube.com/"

    @classmethod
    def site_name(self):
        return "YouTube"

    class Meta:
        verbose_name = "youtube video"


class AudioContentItem(ContentItem):
    duration = models.DurationField(
        help_text="Duration of episode.", blank=True, null=True
    )
    listen_count = models.PositiveIntegerField(
        help_text="Number of episode listens.", blank=True, null=True
    )

    class Meta:
        verbose_name = "episode"


class SpotifyContentItemQuerySet(ContentItemQuerySet):
    assemble_by_ids = assemble_spotify_content_items
    url_converters = ((SpotifyEpisodeURLConverter(), "item_id"),)


class SpotifyContentItem(AudioContentItem):
    objects = SpotifyContentItemQuerySet.as_manager()
    item_id = models.CharField(
        "episode ID", max_length=30, unique=True, help_text="Episode ID."
    )
    sp_show_id = models.CharField("show ID", max_length=30, help_text="Show ID.")
    sp_show_title = models.CharField(
        "show title", max_length=200, help_text="Show title."
    )
    sp_description = models.TextField(
        "description", max_length=5000, blank=True, help_text="Episode description."
    )

    @property
    def content_url(self):
        return SpotifyEpisodeURLConverter().to_url(self.item_id)

    def get_absolute_url(self):
        return reverse("spotifycontentitem_detail", kwargs={"item_id": self.item_id})

    @property
    def site_url(self):
        return "https://open.spotify.com/"

    @classmethod
    def site_name(self):
        return "Spotify"

    class Meta:
        verbose_name = "spotify episode"


class TextContentItem(ContentItem):
    word_count = models.PositiveIntegerField(
        blank=True, null=True, help_text="Word count."
    )
    text_html = models.TextField(
        "content text HTML", blank=True, help_text="Content text HTML."
    )
    text_plain = models.TextField(
        "content plaintext", blank=True, help_text="Content plaintext."
    )


class OBContentItemQuerySet(ContentItemQuerySet):
    assemble_by_ids = assemble_ob_content_items
    url_converters = (
        (OBPostLongURLConverter(), "item_id"),
        (OBPostShortURLConverter(), "ob_post_number"),
    )

    def download_new_items(self, min_publish_date=None):
        """Add posts whose names are not found in the database.

        Do not return items which were not successfully created.
        """
        if min_publish_date is None:
            try:
                # Ignore names of posts published before first download date
                min_publish_date = self.earliest(
                    "download_timestamp"
                ).download_timestamp
            except self.model.DoesNotExist:
                # No items
                pass

        site_names = [
            name
            for name, date in assemble_ob_edit_dates().items()
            if min_publish_date is None or date > min_publish_date
        ]
        db_names = self.values_list("item_id", flat=True)
        names_to_add = [name for name in site_names if name not in db_names]

        created_items = self.create_items(names_to_add)
        return [item for item in created_items if item is not None]

    def update_edited_items(self):
        """Update posts with unsaved edits."""
        self.update_last_edit_dates()
        items_for_update = self.filter(edit_date__gte=F("download_timestamp"))
        updated_items = items_for_update.update_items()
        return [result[0] for result in updated_items]

    def update_last_edit_dates(self):
        """Synchronise edit dates with the overcomingbias site."""
        edit_dates = assemble_ob_edit_dates()

        all_items = self.all()
        for item in all_items:
            item.edit_date = edit_dates[item.item_id]
        update_count = self.bulk_update(all_items, ["edit_date"], batch_size=1000)

        return update_count


class OBContentItem(TextContentItem):
    objects = OBContentItemQuerySet.as_manager()
    item_id = models.CharField(
        "string ID",
        max_length=100,
        unique=True,
        help_text='Post string identifier. E.g. "2006/11/introduction"',
    )
    ob_post_number = models.PositiveIntegerField(
        "number ID", unique=True, help_text="Post number identifier."
    )
    disqus_id = models.CharField(
        max_length=200, blank=True, help_text="Post Disqus API string identifier."
    )
    ob_likes = models.PositiveSmallIntegerField(
        "likes", blank=True, null=True, help_text="Number of likes."
    )
    ob_comments = models.PositiveSmallIntegerField(
        "comments", blank=True, null=True, help_text="Number of comments."
    )

    @property
    def content_url(self):
        return OBPostLongURLConverter().to_url(self.item_id)

    def get_absolute_url(self):
        return reverse("obcontentitem_detail", kwargs={"item_id": self.item_id})

    @property
    def site_url(self):
        return "https://www.overcomingbias.com/"

    @classmethod
    def site_name(self):
        return "OvercomingBias"

    class Meta:
        verbose_name = "overcomingbias post"



from django.contrib import admin, messages
from django.contrib.admin.helpers import Fieldset
from django.contrib.admin.templatetags.admin_urls import add_preserved_filters
from django.core.exceptions import PermissionDenied, SuspiciousOperation
from django.db import IntegrityError
from django.forms import BaseInlineFormSet
from django.http import Http404, HttpResponseRedirect
from django.template.response import TemplateResponse
from django.urls import path, reverse
from ordered_model.admin import OrderedInlineModelAdminMixin, OrderedTabularInline

from obapi import utils
from obapi.exceptions import APICallError
from obapi.forms import (
    AddOBContentItemForm,
    AddSpotifyContentItemForm,
    AddYoutubeContentItemForm,
)
from obapi.models import (
    Author,
    AuthorAlias,
    ContentItem,
    Idea,
    IdeaAlias,
    OBContentItem,
    Sequence,
    SequenceMember,
    SpotifyContentItem,
    Tag,
    TagAlias,
    Topic,
    TopicAlias,
    YoutubeContentItem,
)
from obapi.models.classifiers import CLASSIFIER_SLUG_MAX_LENGTH, ExternalLink

# Inlines
# https://docs.djangoproject.com/en/4.0/ref/contrib/admin/#inlinemodeladmin-objects


class AliasInlineFormset(BaseInlineFormSet):
    def clean(self):
        super().clean()
        # Ensure aliases differ from the name
        name_slug = utils.to_slug(
            self.data["name"], max_length=CLASSIFIER_SLUG_MAX_LENGTH
        )
        forms_to_delete = self.deleted_forms
        valid_forms = [
            form
            for form in self.forms
            if form.is_valid() and form not in forms_to_delete
        ]

        for form in valid_forms:
            # If an alias matches the current name, mark its form for deletion
            alias = form.cleaned_data["text"]
            if alias == name_slug:
                form.cleaned_data["DELETE"] = True


class AliasInline(admin.TabularInline):
    extra = 0
    formset = AliasInlineFormset

    def get_queryset(self, request):
        return super().get_queryset(request).filter(protected=False)


class AuthorAliasInline(AliasInline):
    model = AuthorAlias


class IdeaAliasInline(AliasInline):
    model = IdeaAlias


class TopicAliasInline(AliasInline):
    model = TopicAlias


class TagAliasInline(AliasInline):
    model = TagAlias


class ContentItemInline(admin.StackedInline):
    model = ContentItem.external_links.through
    extra = 0

    autocomplete_fields = ("contentitem",)


# Admin actions
# https://docs.djangoproject.com/en/4.0/ref/contrib/admin/actions/


@admin.action(description="Merge objects", permissions=["change"])
def merge_aliased_objects(modeladmin, request, queryset):
    """Merge multiple objects and their aliases."""
    if (object_count := queryset.count()) < 2:
        modeladmin.message_user(
            request,
            "Two or more items must be selected to merge. No items have been changed.",
            messages.WARNING,
        )
        return

    new_object = queryset.merge_objects()

    modeladmin.message_user(
        request,
        f"{object_count} objects merged into {new_object.name}.",
        messages.SUCCESS,
    )


def convert_aliased_objects(modeladmin, request, queryset, model):
    # Track errors and successes
    err_count = success_count = 0

    # Attempt to convert each object in the queryset
    for obj in queryset:
        try:
            obj.convert_object(model)
            success_count = success_count + 1
        except IntegrityError:
            err_count = err_count + 1
            pass

    # Message user about errors and successes
    if success_count > 0:
        modeladmin.message_user(
            request,
            f"{success_count} {model._meta.verbose_name_plural} successfully created",
            messages.SUCCESS,
        )
    if err_count > 0:
        modeladmin.message_user(
            request,
            f"{err_count} {model._meta.verbose_name_plural} could not be created",
            messages.WARNING,
        )


@admin.action(description="Convert selected items to Authors")
def convert_to_authors(modeladmin, request, queryset):
    convert_aliased_objects(modeladmin, request, queryset, Author)


@admin.action(description="Convert selected items to Ideas")
def convert_to_ideas(modeladmin, request, queryset):
    convert_aliased_objects(modeladmin, request, queryset, Idea)


@admin.action(description="Convert selected items to Topics")
def convert_to_topics(modeladmin, request, queryset):
    convert_aliased_objects(modeladmin, request, queryset, Topic)


@admin.action(description="Convert selected items to Tags")
def convert_to_tags(modeladmin, request, queryset):
    convert_aliased_objects(modeladmin, request, queryset, Tag)


# Model Admins
# https://docs.djangoproject.com/en/4.0/ref/contrib/admin/#modeladmin-objects


@admin.register(Author)
class AuthorAdmin(admin.ModelAdmin):
    inlines = [AuthorAliasInline]
    actions = [
        merge_aliased_objects,
        convert_to_ideas,
        convert_to_topics,
        convert_to_tags,
    ]
    search_fields = ("name",)


@admin.register(Idea)
class IdeaAdmin(admin.ModelAdmin):
    inlines = [IdeaAliasInline]
    actions = [
        merge_aliased_objects,
        convert_to_authors,
        convert_to_topics,
        convert_to_tags,
    ]
    search_fields = ("name",)


@admin.register(Topic)
class TopicAdmin(admin.ModelAdmin):
    inlines = [TopicAliasInline]
    actions = [
        merge_aliased_objects,
        convert_to_authors,
        convert_to_ideas,
        convert_to_tags,
    ]
    search_fields = ("name",)


@admin.register(Tag)
class TagAdmin(admin.ModelAdmin):
    inlines = [TagAliasInline]
    actions = [
        merge_aliased_objects,
        convert_to_authors,
        convert_to_ideas,
        convert_to_topics,
    ]
    search_fields = ("name",)


@admin.register(ExternalLink)
class ExternalLinkAdmin(admin.ModelAdmin):
    inlines = [
        ContentItemInline,
    ]
    exclude = ("external_links",)
    search_fields = ("url",)
    actions = ("delete_unreferenced_links",)

    @admin.action(description="Delete unreferenced links", permissions=["delete"])
    def delete_unreferenced_links(self, request, queryset):
        delete_count, details = queryset.filter(content__isnull=True).delete()
        self.message_user(
            request,
            f"{delete_count} links deleted",
            messages.INFO,
        )


class ContentItemAdminTemplate(admin.ModelAdmin):
    add_form_template = "admin/obapi/add_form.html"
    AddForm = None  # override in subclasses

    actions = ["update_selected_items", "internalize_links"]

    readonly_fields = ("create_timestamp", "update_timestamp", "download_timestamp")

    search_fields = ("title",)

    autocomplete_fields = (
        "authors",
        "ideas",
        "topics",
        "tags",
        "external_links",
        "internal_links",
    )

    def add_view(self, request, form_url="", extra_context=None):
        if not self.has_add_permission(request):
            raise PermissionDenied
        if request.method == "POST":
            form = self.AddForm(request.POST)
            if form.is_valid():
                id_field = next(iter(form.fields.values())).label
                message_if_error = (
                    f"Could not retrieve item. Please check the {id_field} is valid."
                )
                try:
                    item = form.save()
                except APICallError:
                    item = None
                except IntegrityError:
                    item = None
                    message_if_error = "The specified item already exists."
                if item is not None:
                    change_message = f"Created item {item}."
                    self.log_addition(request, item, change_message)
                    return self.response_add(request, item)

                self.message_user(
                    request,
                    message_if_error,
                    messages.ERROR,
                )

        else:
            form = self.AddForm()

        opts = self.model._meta
        fieldset = Fieldset(form, fields=form.fields, model_admin=self)
        context = {
            **self.admin_site.each_context(request),
            "title": f"Add {self.model._meta.verbose_name}",
            "fieldset": fieldset,
            "media": self.media,
            "errors": form.errors,
            "add": True,
            "has_view_permission": self.has_view_permission(request, None),
            "has_add_permission": self.has_add_permission(request),
            "has_change_permission": self.has_change_permission(request, None),
            "has_delete_permission": self.has_delete_permission(request, None),
            "opts": opts,
            "app_label": opts.app_label,
            **(extra_context or {}),
        }
        request.current_app = self.admin_site.name
        form_template = self.add_form_template
        return TemplateResponse(request, form_template, context)

    @admin.action(description="Update selected items", permissions=["change"])
    def update_selected_items(self, request, queryset):
        queryset.update_items()

    @admin.action(description="Internalize links", permissions=["change"])
    def internalize_links(self, request, queryset):
        try:
            for item in queryset:
                item.internalize_links(clear=False)
        except IntegrityError:
            self.message_user(
                request, "Error when internalizing links.", messages.ERROR
            )
        else:
            self.message_user(
                request, "Links internalized successfully.", messages.SUCCESS
            )


@admin.register(ContentItem)
class ContentItemAdmin(ContentItemAdminTemplate):
    # Revert to standard add view
    def add_view(self, request, form_url="", extra_context=None):
        return self.changeform_view(request, None, form_url, extra_context)

    @admin.action(description="Update selected items", permissions=["change"])
    def update_selected_items(self, request, queryset):
        for model in (YoutubeContentItem, SpotifyContentItem, OBContentItem):
            model.objects.filter(
                pk__in=queryset.select_subclasses(model)
            ).update_items()


@admin.register(YoutubeContentItem)
class YoutubeContentItemAdmin(ContentItemAdminTemplate):
    AddForm = AddYoutubeContentItemForm


@admin.register(SpotifyContentItem)
class SpotifyContentItemAdmin(ContentItemAdminTemplate):
    AddForm = AddSpotifyContentItemForm


@admin.register(OBContentItem)
class OBContentItemAdmin(ContentItemAdminTemplate):
    AddForm = AddOBContentItemForm

    def get_urls(self):
        urls = super().get_urls()
        custom_urls = [
            path(
                "update/",
                self.admin_site.admin_view(self.update_view),
                name="obapi_obcontentitem_update",
            )
        ]
        return custom_urls + urls

    def changelist_view(self, request, extra_context=None):
        if extra_context is None:
            extra_context = {}
        extra_context.update(has_change_permission=self.has_change_permission(request))
        return super().changelist_view(request, extra_context)

    def update_view(self, request):
        """View to update existing OB content items.

        Either downloads all new posts ("pull") or updates all edited posts ("sync").

        Works by (1) performing the requested operation, (2) logging the results and
        messaging the user, and (3) redirecting to list / index page.
        """
        # Accept POST requests only
        if request.method != "POST":
            raise Http404

        if "_pull" in request.POST:
            # Check permissions
            if not self.has_add_permission(request):
                raise PermissionDenied

            self.pull(request)
        elif "_sync" in request.POST:
            # Check permissions
            if not self.has_change_permission(request):
                raise PermissionDenied

            self.sync(request)
        else:
            raise SuspiciousOperation

        # Redirect logic - copied from internal method `_response_post_save`
        opts = self.model._meta
        if self.has_view_or_change_permission(request):
            post_url = reverse(
                "admin:%s_%s_changelist" % (opts.app_label, opts.model_name),
                current_app=self.admin_site.name,
            )
            preserved_filters = self.get_preserved_filters(request)
            post_url = add_preserved_filters(
                {"preserved_filters": preserved_filters, "opts": opts}, post_url
            )
        else:
            post_url = reverse("admin:index", current_app=self.admin_site.name)
        return HttpResponseRedirect(post_url)

    def pull(self, request):
        """Download new posts."""
        # Pull items
        created_items = OBContentItem.objects.download_new_items()
        # Log additions
        for item in created_items:
            self.log_addition(request, item, f"Created item {item}.")
        # Message user
        changecount = len(created_items)
        if changecount == 0:
            msg = "No items were created."
            status = messages.INFO
        else:
            msg = f"Successfully created {changecount} item(s)."
            status = messages.SUCCESS
        self.message_user(request, msg, status)
        return changecount

    def sync(self, request):
        """Updated existing posts."""
        # Sync items
        updated_items = OBContentItem.objects.update_edited_items()
        # Log changes
        for item in updated_items:
            self.log_change(request, item, f"Updated item {item}.")
        # Message user
        changecount = len(updated_items)
        if changecount == 0:
            msg = "No items were updated."
            status = messages.INFO
        else:
            msg = f"Successfully updated {changecount} item(s)."
            status = messages.SUCCESS
        self.message_user(request, msg, status)
        return changecount


class SequenceMemberInline(OrderedTabularInline):
    model = SequenceMember
    fields = (
        "content_item",
        "order",
        "move_up_down_links",
    )
    readonly_fields = (
        "order",
        "move_up_down_links",
    )
    ordering = ("order",)
    extra = 1
    autocomplete_fields = ("content_item",)


@admin.register(Sequence)
class SequenceAdmin(OrderedInlineModelAdminMixin, admin.ModelAdmin):
    model = Sequence
    list_display = ("title",)
    inlines = (SequenceMemberInline,)



"""Download raw API data from the web."""

from base64 import b64encode

import cachetools.func
import httpx
from django.conf import settings
from obscraper import get_edit_dates, get_posts_by_names

from obapi import exceptions


def download_youtube_videos_json(video_ids):
    api_url = "https://youtube.googleapis.com/youtube/v3/videos"
    payload = {
        "id": ",".join(video_ids),
        "part": ("snippet", "contentDetails", "statistics"),
        "key": settings.YOUTUBE_API_KEY,
    }
    response = httpx.get(url=api_url, params=payload)
    _raise_for_status(response)
    return response.json()


def download_spotify_episodes_json(episode_ids):
    api_url = "https://api.spotify.com/v1/episodes/"
    headers = {"Authorization": f"Bearer {_get_spotify_api_token()}"}
    params = {"ids": ",".join(episode_ids), "market": "US"}
    response = httpx.get(url=api_url, headers=headers, params=params)
    _raise_for_status(response)
    return response.json()


def download_ob_post_objects(post_names):
    try:
        post_dict = get_posts_by_names(post_names)
    except ValueError as err:
        raise exceptions.APICallError("API call failed.") from err
    else:
        return post_dict


def download_ob_edit_dates():
    return get_edit_dates()


@cachetools.func.ttl_cache(ttl=3)
def _get_spotify_api_token():
    auth_url = "https://accounts.spotify.com/api/token"
    auth_header = b64encode(
        f"{settings.SPOTIFY_CLIENT_ID}:{settings.SPOTIFY_CLIENT_SECRET}".encode("ascii")
    ).decode("ascii")
    headers = {"Authorization": f"Basic {auth_header}"}
    payload = {"grant_type": "client_credentials"}
    response = httpx.post(url=auth_url, data=payload, headers=headers)
    _raise_for_status(response)
    token_info = response.json()
    return token_info["access_token"]


def _raise_for_status(response):
    try:
        response.raise_for_status()
    except httpx.HTTPError as err:
        raise exceptions.APICallError("API call failed.") from err



import datetime
import re

import bs4
from dateutil.parser import isoparse

from obapi import utils


def tidy_youtube_videos_json(video_ids, raw_json):
    items = raw_json["items"]

    # Set item to None if not present
    video_dict = {item["id"]: item for item in items}
    for video_id in video_ids:
        video_dict.setdefault(video_id, None)
    videos = [video_dict[video_id] for video_id in video_ids]

    return [_tidy_youtube_video_json(video) for video in videos]


def tidy_spotify_episodes_json(episode_ids, raw_json):
    items = raw_json["episodes"]

    # Set item to None if not present
    episode_dict = {item["id"]: item for item in items}
    for episode_id in episode_ids:
        episode_dict.setdefault(episode_id, None)
    episodes = [episode_dict[episode_id] for episode_id in episode_ids]

    return [_tidy_spotify_episode_json(episode) for episode in episodes]


def tidy_ob_post_objects(post_names, post_dict):
    posts = [post_dict[post_name] for post_name in post_names]
    return [_tidy_ob_post_object(post) for post in posts]


def _tidy_youtube_video_json(item_json):
    """Tidy a YouTube API v3 response into a dictionary."""
    if item_json is None:
        return None
    video = {
        "author_names": [item_json["snippet"]["channelTitle"]],
        "classifier_names": item_json["snippet"].get("tags", []),
        "title": item_json["snippet"]["title"],
        "description_html": utils.plaintext_to_html(
            item_json["snippet"].get("description", "")
        ),
        "publish_date": isoparse(item_json["snippet"]["publishedAt"]),
        "duration": utils.parse_duration(item_json["contentDetails"]["duration"]),
        "view_count": int(item_json["statistics"]["viewCount"]),
        "item_id": item_json["id"],
        "yt_channel_id": item_json["snippet"]["channelId"],
        "yt_channel_title": item_json["snippet"]["channelTitle"],
        "yt_likes": item_json["statistics"].get("likeCount", None),
        "yt_description": item_json["snippet"].get("description", ""),
    }
    return video


def _tidy_spotify_episode_json(item_json):
    if item_json is None:
        return None
    episode = {
        "author_names": [item_json["show"]["publisher"]],
        "title": item_json["name"],
        "description_html": item_json["html_description"],
        "publish_date": isoparse(item_json["release_date"]).replace(
            tzinfo=datetime.timezone.utc
        ),
        "duration": datetime.timedelta(milliseconds=item_json["duration_ms"]),
        "item_id": item_json["id"],
        "sp_show_id": item_json["show"]["id"],
        "sp_show_title": item_json["show"]["name"],
        "sp_description": item_json["description"],
    }
    return episode


def _tidy_ob_post_object(item_post):
    if item_post is None:
        return None
    internal_links = [_tidy_ob_internal_link(link) for link in item_post.internal_links]
    text_html = _tidy_ob_post_html(item_post.text_html)
    disqus_id = item_post.disqus_id or ""
    post = {
        "author_names": [item_post.author],
        "classifier_names": [*item_post.tags, *item_post.categories],
        "link_urls": [*internal_links, *item_post.external_links],
        "title": item_post.title,
        "publish_date": item_post.publish_date,
        "edit_date": item_post.edit_date,
        "word_count": item_post.word_count,
        "text_html": text_html,
        "text_plain": item_post.plaintext,
        "item_id": item_post.name,
        "ob_post_number": item_post.number,
        "disqus_id": disqus_id,
        "ob_likes": item_post.votes,
        "ob_comments": item_post.comments,
    }
    return post


def _tidy_ob_internal_link(url: str):
    """Fix common errors with internal links."""
    url = url.strip()  # strip whitespace
    url = re.sub(r"\.htm(?!l)", ".html", url)  # ensure .html not .htm
    url = re.sub(r"-+", "-", url)  # collapse duplicate hyphens
    url = re.sub(r"-\.html", ".htm", url)  # remove trailing hyphens
    return url


def _tidy_ob_post_html(text_html: str):
    # Replace nbsp
    text = text_html.replace("\xa0", " ")

    # Replace multiple spaces?
    # re.sub(" {2,}", " ", text)

    soup = bs4.BeautifulSoup(text, "lxml")

    # Unwrap unwanted tags
    TAGS_TO_UNWRAP = ("nobr", "span")
    for tag in soup.find_all(TAGS_TO_UNWRAP):
        tag.unwrap()

    def replace_tag(soup, filter, replacement, clear_attributes=False):
        """Replace all tags matching a filter."""
        for tag in soup.find_all(filter):
            tag.name = replacement
            if clear_attributes:
                tag.attrs = {}

    # Replace data-list items
    replace_tag(soup, ["dt", "dd"], "p", clear_attributes=True)
    replace_tag(soup, "dl", "blockquote", clear_attributes=True)

    # Apply single block-quotes
    def is_single_blockquote(tag):
        if tag.name not in ["p", "div"]:
            return False

        if class_ := tag.attrs.get("class"):
            if "blockquote" in class_:
                return True

        bq_styles = ["margin-left: 40px", "padding-left: 30px", "padding-left: 40px"]
        if style := tag.attrs.get("style"):
            if any(bq_style in style for bq_style in bq_styles):
                return True

        return False

    for tag in soup.find_all(is_single_blockquote):
        tag.wrap(soup.new_tag("blockquote"))
        tag.name = "p"
        tag.attrs = {}

    # Double blockquotes
    def is_double_blockquote(tag):
        if tag.name != "p":
            return False

        bq_styles = ["padding-left: 60px;"]
        if style := tag.attrs.get("style"):
            if any(bq_style in style for bq_style in bq_styles):
                return True

        return False

    for tag in soup.find_all(is_double_blockquote):
        tag.wrap(soup.new_tag("blockquote"))
        tag.wrap(soup.new_tag("blockquote"))
        tag.name = "p"
        tag.attrs = {}

    # Remove class attributes
    for tag in soup.find_all(class_="MsoNormal"):
        del tag["class"]

    # Extract fragment within entry-content div
    html_fragment = soup.find(class_="entry-content").decode_contents().strip()
    return html_fragment



import random

import obscraper
import pytest
from obapi.models import OBContentItem
from obapi.models.content import SpotifyContentItem, YoutubeContentItem

from sample_data import SAMPLE_SPOTIFY_IDS, SAMPLE_YOUTUBE_IDS


def create_random_items(sample_ids, n, model):
    sample_ids = random.sample(sample_ids, n)
    created_items = model.objects.create_items(sample_ids)
    return [item for item in created_items if item is not None]


@pytest.fixture(scope="session")
def obcontent_edit_dates():
    return obscraper.get_edit_dates()


@pytest.fixture(scope="session")
def random_obcontentitems(obcontent_edit_dates):
    """Factory function which populates the database with OvercomingBias posts."""

    def _random_obcontentitems(n):
        sample_ids = sorted(obcontent_edit_dates.keys())
        return create_random_items(sample_ids, n, OBContentItem)

    return _random_obcontentitems


@pytest.fixture(scope="session")
def random_youtubecontentitems():
    """Factory function which populates the database with YouTube videos."""

    def _random_youtubecontentitems(n):
        return create_random_items(SAMPLE_YOUTUBE_IDS, n, YoutubeContentItem)

    return _random_youtubecontentitems


@pytest.fixture(scope="session")
def random_spotifycontentitems():
    """Factory function which populates the database with Spotify episodes."""

    def _random_spotifycontentitems(n):
        return create_random_items(SAMPLE_SPOTIFY_IDS, n, SpotifyContentItem)

    return _random_spotifycontentitems



import pytest
from obapi.download import (
    _get_spotify_api_token,
    download_spotify_episodes_json,
    download_youtube_videos_json,
)
from obapi.exceptions import APICallError

from markers import require_spotify_api_auth, require_youtube_api_key


@require_youtube_api_key
class TestDownloadYoutubeVideosJSON:
    def test_no_error_raised_for_invalid_video_ids(self):
        result = download_youtube_videos_json(video_ids=["123"])
        assert isinstance(result, dict)

    def test_raises_error_for_invalid_api_key(self, settings):
        # Arrange
        settings.YOUTUBE_API_KEY = "FakeValue"

        # Act & assert
        with pytest.raises(APICallError):
            download_youtube_videos_json(video_ids=["C-gEQdGVXbk"])

    def test_works_correctly_for_valid_inputs(self):
        # Act
        test_video = download_youtube_videos_json(video_ids=["C-gEQdGVXbk"])

        # Assert
        assert "items" in test_video
        assert (
            test_video["items"][0]["snippet"]["title"]
            == "10 Python Tips and Tricks For Writing Better Code"
        )


@require_spotify_api_auth
class TestDownloadSpotifyEpisodesJSON:
    def test_raises_error_for_invalid_episode_ids(self):
        with pytest.raises(APICallError):
            download_spotify_episodes_json(episode_ids=["123"])

    def test_raises_error_for_invalid_user_credentials(self, settings):
        _get_spotify_api_token.cache_clear()
        settings.SPOTIFY_CLIENT_ID = "fake-id"
        with pytest.raises(APICallError):
            download_spotify_episodes_json(episode_ids=["6MAszRR6tdDnMsjgVdw4Jh"])

        settings.SPOTIFY_CLIENT_SECRET = "d7JxFQUjNMCBn6agfBtWfnM6BcioGowC"
        with pytest.raises(APICallError):
            download_spotify_episodes_json(episode_ids=["6MAszRR6tdDnMsjgVdw4Jh"])

    def test_returns_none_for_nonexistent_episodes(self):
        # Act
        non_episode = download_spotify_episodes_json(["6MAszRR6tdDnMsjgVdw4Jp"])
        # Assert
        assert non_episode["episodes"] == [None]

    def test_works_correctly_for_valid_inputs(self):
        # Act
        test_episode = download_spotify_episodes_json(
            episode_ids=["6MAszRR6tdDnMsjgVdw4Jh"]
        )
        # Assert
        assert "episodes" in test_episode
        assert (
            test_episode["episodes"][0]["show"]["publisher"]
            == "Mercatus Center at George Mason University"
        )



import os

from dotenv import find_dotenv, load_dotenv

# take environment variables from .env
load_dotenv(find_dotenv())

# Applications

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "obapi",
]

# Secrets
SECRET_KEY = os.getenv("DJANGO_SECRET_KEY", "fake-key")
YOUTUBE_API_KEY = os.getenv("YOUTUBE_API_KEY")
SPOTIFY_CLIENT_ID = os.getenv("SPOTIFY_CLIENT_ID")
SPOTIFY_CLIENT_SECRET = os.getenv("SPOTIFY_CLIENT_SECRET")


# Debug mode
DEBUG = False

# Internationalization
# https://docs.djangoproject.com/en/4.0/topics/i18n/


USE_TZ = True

TIME_ZONE = "UTC"

# Database

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": ":memory:",
    }
}

# Templates

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
            ],
        },
    },
]



from django.db import models
from obapi import utils
from obapi.modelfields import SimpleSlugField
from ordered_model.models import OrderedModel

SEQUENCE_SLUG_MAX_LENGTH = 150


class BaseSequence(models.Model):
    """Base class for Sequence models."""

    title = models.CharField(max_length=100, help_text="Sequence title.")
    slug = SimpleSlugField(max_length=SEQUENCE_SLUG_MAX_LENGTH, editable=False)
    abstract = models.TextField(
        blank=True, max_length=5000, help_text="Description of sequence."
    )
    items = models.ManyToManyField("ContentItem", through="BaseSequenceMember")

    def __str__(self):
        return self.title

    class Meta:
        abstract = True

    def clean(self):
        # Set slug from title
        self.slug = utils.to_slug(self.title, max_length=SEQUENCE_SLUG_MAX_LENGTH)
        super().clean()

    def save(self, *args, **kwargs):
        """Save a Sequence object."""
        self.clean()
        super().save(*args, **kwargs)


class BaseSequenceMember(OrderedModel):
    """Base class for Sequence items."""

    order_with_respect_to = "sequence"
    content_item = models.ForeignKey(
        "ContentItem",
        on_delete=models.CASCADE,
        related_name="sequence_members",
        related_query_name="sequence_members",
    )
    sequence = models.ForeignKey(
        BaseSequence,
        on_delete=models.CASCADE,
        related_name="members",
        related_query_name="members",
    )

    class Meta:
        ordering = ("sequence", "order")
        abstract = True

    def __str__(self):
        return f"{self.content_item.title} ({self.sequence.title})"


class Sequence(BaseSequence):
    items = models.ManyToManyField("ContentItem", through="SequenceMember")

    class Meta:
        constraints = [
            models.UniqueConstraint(fields=["slug"], name="unique_sequence_slug")
        ]


class SequenceMember(BaseSequenceMember):
    sequence = models.ForeignKey(
        Sequence,
        on_delete=models.CASCADE,
        related_name="members",
        related_query_name="members",
    )



import datetime

import pytest
from obapi.models import OBContentItem


class TestFindByURL:
    pass


@pytest.mark.django_db
def test_update_last_edit_dates(random_obcontentitems):
    items = random_obcontentitems(5)
    true_edit_dates = [item.edit_date for item in items]
    for item in items:
        item.edit_date = datetime.datetime(2010, 1, 1, tzinfo=datetime.timezone.utc)

    item_ids = [item.item_id for item in items]
    items_queryset = OBContentItem.objects.filter(item_id__in=item_ids)

    items_queryset.update_last_edit_dates()

    # Assert
    for i, item in enumerate(items):
        item.refresh_from_db()
        assert item.edit_date == true_edit_dates[i]



from django.db import models

import obapi.formfields
from obapi import validators


class SimpleSlugField(models.CharField):
    default_validators = [validators.validate_simple_slug]
    description = "Custom slug (up to %(max_length)s)"

    def __init__(self, *args, max_length=50, db_index=True, **kwargs):
        super().__init__(*args, max_length=max_length, db_index=db_index, **kwargs)

    def deconstruct(self):
        name, path, args, kwargs = super().deconstruct()
        if kwargs.get("max_length") == 50:
            del kwargs["max_length"]
        if self.db_index is False:
            kwargs["db_index"] = False
        else:
            del kwargs["db_index"]
        return name, path, args, kwargs

    def get_internal_type(self):
        return "SlugField"

    def formfield(self, **kwargs):
        return super().formfield(
            **{"form_class": obapi.formfields.SimpleSlugField, **kwargs}
        )



import re

from django.core.validators import RegexValidator

validate_simple_slug = RegexValidator(
    re.compile(r"^[-a-z0-9]+\Z"),
    "Enter a valid “slug” consisting of lowercase letters, numbers or hyphens.",
    "invalid",
)



SAMPLE_YOUTUBE_IDS = (
    "V84_F1QWdeU",
    "Urk3xn7l3AM",
    "0lKliaFllPA",
    "izEUig9w1tM",
    "n1DQ6uwfGE4",
    "N4I2If3x_9g",
    "aspMV6ERqpo",
    "9qcIsjrHENU",
    "euarux37A44",
    "VngstyKI754",
    "Gd9GOoITfXY",
    "wC6krCh48-Q",
    "2HN_xz_-pfg",
    "zGXpsJYNILg",
    "Rjm--7t8Llk",
    "jxJJkWHQF7k",
    "IBc1oVXen-o",
    "9Ix0o2iEcww",
    "4iMKd7LtgHQ",
    "YGDpfI7I04Y",
    "3tcYQE1B588",
    "vdV_PjTEaaY",
    "aumhEQtNyF4",
    "2Pe9ZLmvoDU",
    "4yZKGbq1YmA",
    "qmGOhT-8rJ4",
)

SAMPLE_SPOTIFY_IDS = (
    "3KLwjmeHEOg9x2ZfhbMUnW",
    "6z8lvJia3OzcIVCStbmEtU",
    "4eMenBlRap0b6FKsG6TWYg",
    "1LUh8e5N9lY7QJWi6ahJCo",
    "6MAszRR6tdDnMsjgVdw4Jh",
    "6dXQ2l97epGGwmdlBJQgj2",
    "7hzVvRkxjTwfR8Nlg1SvRR",
    "76wXKfIW9iwXtAUSkrf6xY",
    "7ufnUibHWOvxb8ROD4eALT",
    "5BkIWk6VYQCTZOQaRN4xa5",
    "0pgNYJjZFPlLssOzOVEdGa",
    "3LEhJxIYhjEL8oLjuqRXPE",
    "3hhZSVKfsAxqo4SphOKLqZ",
    "7acAKF2oEAjkzlYU8sRvbm",
    "7ipidZiUFsy1GVXpsoJGEC",
    "1GSUK92hQqWdOMNzs5j1IK",
    "5bHYUb0Gd1SFjmuhLzCcOQ",
    "6jqwwyNmeKzhU5B5dHoYsv",
    "2lRH5Y2FqjT2KZGW6cdP8R",
)
